#!/usr/bin/env python3

import multiprocessing as mp
import pandas as pd
import numpy as np
from itertools import repeat
from numba import njit
import hashlib
import re
import scipy.stats as stats

@njit
def table_lookup(coord,data):
    """
    njit compiled np.where to speed up computation

    Parameters
    ----------
    coord : encoding genomic coordinate, barcode and 
    transposon orientation
    data : numpy array
        all ids

    Returns
    -------
    numpy array
        indices of coord in data

    """
    return np.where(data==coord)[0]



def get_filter_mask(id_, all_ids, data):
    """
    Get filter mask to remove all barcodes that have count sums below 
    10 % of the max barcode.

    Parameters
    ----------
    id_ : encoding genomic coordinate, barcode and 
    transposon orientation
    all_ids : numpy array
        all hashed ids encoding genomic coordinate, barcode and 
        transposon orientation
    data : numpy array
        count matrix

    Returns
    -------
    boolean array
        holding the filter mask

    """
    sub = data[table_lookup(id_,all_ids),:]
    
    total = np.sum(sub)
    rowsums = np.sum(sub, axis=1)
    row_fracs = rowsums/total

    # cutoff = stats.norm.ppf(0.0001, loc=np.mean(rowsums), scale=np.mean(rowsums)/4)
    # return rowsums > cutoff
    return row_fracs >= 0.1

def filter_barcodes(data,experiment):
    """
    Filter out barcodes with little read counts.

    Parameters
    ----------
    data : pandas dataframe
        count matrix
    experiment : LibraryExperiment
        Object holding input parameters.

    Returns
    -------
    out : pandas dataframe
        Filtered count matrix

    """
    out= data.copy()
    pools = experiment.pools
    out["id"] = out["Feature"].astype(str) + out["Orientation"]
    ids = get_hash_ids(out["id"]) 
    out["id"] = ids
    assert len(np.unique(ids)) == len(np.unique(out["id"])), "Hashed ids not unique!"

    unique_ids = out["id"].unique()
        
    pool = mp.Pool(experiment.cores)
    mask = pool.starmap_async(get_filter_mask, zip(unique_ids, repeat(ids),
                              repeat(out[pools].values.astype(int)))).get()
    pool.close()
    
    mask = np.hstack(mask)
    out = out[mask]
    out = out.drop("id", axis=1)
    return out

def filter_counts(counts, thr=0.05):
    """
    Apply local relative filter. Counts are set to zero if their ratio to the 
    max count is < thr.

    Parameters
    ----------
    counts : numpy array
        count matrix
    thr : int, optional
        relative filter threshold. The default is 0.05.

    Returns
    -------
    counts : numpy array
        filtered count matrix

    """
    m = np.max(counts, axis=1)
    mask = (counts.T/m).T < thr
    counts[mask] = 0
    
    return counts


def get_pool_table_line(id_, bowtie_res, all_ids, pools, poolidx_dict):
    """
    

    Parameters
    ----------
    id_ : int
        hashed ids encoding genomic coordinate, barcode and 
        transposon orientation
    bowtie_res : pandas dataframe
        parsed output of bowtie2, each line corresponds to a read
    all_ids : numpy array
        all unique hashed ids encoding genomic coordinate, barcode and 
        transposon orientation
    pools : list of strings
       all pools
    poolidx_dict : dict
        of poolnames to indices

    Returns
    -------
    line : numpy array
       line of count matrix for mutant id
    line_id : int
        hashed id encoding genomic coordinate, barcode and 
        transposon orientation

    """
    coord_subset = bowtie_res[table_lookup(id_, all_ids)]
    
    # pools are in column 1
    poolcounts = np.unique(coord_subset[:,1], return_counts=True)
    line_id = np.unique(coord_subset[:,5])[0]
    line = np.zeros((1,len(pools)), dtype=int)
    for j in range(len(poolcounts[0])):
        line[0,poolidx_dict[poolcounts[0][j]]] = poolcounts[1][j]

    return line, line_id


def get_count_matrix(ids, bowtie_res, experiment):
    """
    Assembles count matrix from bowtie2 output.

    Parameters
    ----------
    ids : int
        hashed ids encoding genomic coordinate, barcode and 
        transposon orientation
    bowtie_res : pandas dataframe
        parsed output of bowtie2, each line corresponds to a read
    experiment : LibraryExperiment
        Object holding input parameters.

    Returns
    -------
    out : pandas dataframe
        Mutant count matrix

    """
    # get all pools
    pools = experiment.pools
    
    # make sure pools are always on same position
    poolidx_dict = {p : i for i,p in enumerate(pools)}
    
    unique_ids, counts = np.unique(ids, return_counts=True)
    # exclude barcodes which occur only once
    unique_ids = unique_ids[counts>=5]
    print("Generating count matrix with", len(unique_ids), "mutants")

    pool = mp.Pool(experiment.cores)
    pool_table_lines,line_ids = zip(*pool.starmap_async(get_pool_table_line, 
                                          zip(unique_ids, 
                                              repeat(bowtie_res.values), 
                                              repeat(ids),
                                              repeat(pools), 
                                              repeat(poolidx_dict))
                                          ).get())

    pool.close()
    pool_presence_table = np.vstack(pool_table_lines)
    table_ids  = np.hstack(line_ids)
    orient = [i[0] for i in table_ids]
    coord_match  = re.compile(r'(\d+)(\D+)')
    coord = []
    bar = []
    for i in table_ids:        
        m = coord_match.search(i)
        coord.append(m.group(1))
        bar.append(m.group(2))
    
    coord = np.array(coord, dtype=int)
    orient = np.array(orient)
    bar = np.array(bar)
    out_ids = np.vstack((coord, orient, bar)).T
    out = pd.DataFrame(np.hstack((out_ids,pool_presence_table)), columns = ["Feature","Orientation", "Barcode"] + list(pools) )
    out["Feature"] = out["Feature"].astype(int)
    # important to sort for barcode filtering
    out = out.sort_values(["Feature", "Orientation"], axis=0)
    out = out.reset_index(drop=True)
    print("Generated count matrix!")
 
    return out


def get_hash_ids(ids):
    """
    Concatenate orientation, coordinate and barcode to create unique ids
    Return unique int identifiers along with original str ids

    Parameters
    ----------
    ids : list of str
        Ids indicating genomic coordinate, barcode and Tn orientation of mutant

    Returns
    -------
    encoded_id : list of int
        hashed id

    """

    # generate unique int id by creating a hash from the string and convert to int
    # numba only accepts int or bools
    encoded_id = []
    for i in ids:
        m = hashlib.md5()
        m.update(i.encode("utf8"))
        encoded_id.append(str(int(m.hexdigest(), 16))[0:16])
    encoded_id = np.array(encoded_id, dtype=int)
    
    return encoded_id

    
    


    
    


#!/usr/bin/env python3

import subprocess
import os

from pysudoku_package.io import txt_writer


def parse_bowtie2_output(experiment):
    """
    Parse output of bowtie2 and write to alignment_result.csv in 
    temp directory.

    Parameters
    ----------
    experiment : LibraryExperiment
        Object holding input parameters.

    Returns
    -------
    None.

    """
    flag_list = [0, 16]
    # if param['seq_type']=="PE": #paired ended
    #     flag_list = [83, 99] 

    out=[]
    outfile = os.path.join("temp", "alignment_result.csv")
    align_res = experiment.alignment
    with open(align_res) as current:
        for line in current:
            line = line.split('\t')
            if (line[0][0] != "@") and (line[0][0] != ">"): #ignores headers and unaligned contigs
                id_str = line[0]
                flag_sum = int(line[1])
                coord = int(line[3])
                quality = int(line[4])
                cigar = line[5]
                multi = "XS:i:" in line
                orientation = '+'
                if flag_sum == flag_list[1]:
                    orientation = '-'
                    # considers CIGAR scores. S means that the read is clipped, so the position will actually be different
                    # do we need that in end to end mode?
                    # matches = findall(r'(\d+)([A-Z]{1})', cigar)
                    # clipped = 0
                    # for match in matches:
                    #     if match[1] == "S":
                    #         clipped=int(match[0])
                    #         break #only consideres the first one at the start
                    # coord = coord + len(line[9]) - clipped
                    # subtract 2 if tn is in reverse direction
                    coord = coord + len(line[9]) - 2
                if (flag_sum in flag_list) & (multi==False) & (quality >= experiment.map_quality):
                    read_id,pool,barcode = id_str.split(":")
                    out.append(f"{read_id},{pool},{coord},{orientation},{barcode}\n")
            #if psutil.virtual_memory().percent>=80:
            if len(out)>500000:
                txt_writer(outfile, out)
                out=[]
                        
    txt_writer(outfile, out)


def run_bowtie2(experiment):
    """
    Calls bowtie2 alignment from subprocess and runs alignment on 
    trimmed_sequences.fastq. Bowtie2 writes alignments to temp/aligment.sam.

    Parameters
    ----------
    experiment : LibraryExperiment
        Object holding input parameters.


    -------
    None.

    """

    out = experiment.alignment
    seq = os.path.join("temp", "trimmed_sequences.fastq")
    subprocess.call(['bowtie2', 
                     '-x', 
                     experiment.bowtie_ref, 
                     '-q', 
                     seq, 
                     "--end-to-end", 
                     f"-S {out}",
                     '--no-unal',
                     '--threads', 
                     str(experiment.cores)])




#!/usr/bin/env python3

import pandas as pd
import numpy as np
from itertools import product
from Bio import SeqIO
import math
import functools

def get_ambiguity(experiment):
    """
    Finds ambiguous and unambiguous mutants and returns them as separate arrays

    Parameters
    ----------
    experiment : LibraryExperiment
        Object holding input parameters.


    Returns
    -------
    unambiguous_data : numpy array
    ambiguous_data : numpy array

    """    
    

    # get only columns with read counts, so should also work with the pool_presence_table containing reads.
  
    dims = experiment.pool_dims
    data = experiment.count_mat
    
    pool_reads = data[experiment.pools].astype(int)
    
    nonzero_dimcounts = []
    for i in dims:
        nonzero_dimcounts.append(np.sum(pool_reads[dims[i]]>0, axis=1))
    
    dim_prod = functools.reduce(lambda x,y: x*y, nonzero_dimcounts)
    
    unpredictable  = dim_prod == 0
    
    nonzero_dimcounts = np.vstack(nonzero_dimcounts).T
    
    ambiguous =np.sum(nonzero_dimcounts > 1, axis=1) >=2
    ambiguous[unpredictable] = False
    ambiguous_data = data.loc[ambiguous,:]

    # we only allow one axis with more than 1 nonzero entry for unambiguous mutants
    unambiguous =np.sum(nonzero_dimcounts > 1, axis=1) <= 1
    unambiguous[unpredictable] = False
    unambiguous_data = data.loc[unambiguous,:]
    
    return unambiguous_data, ambiguous_data
    
def get_unambiguous_locations(data, experiment):
    """
    

    Parameters
    ----------
    data : pandas dataframe
        count matrix of unambiguous mutants
    experiment :  LibraryExperiment
        Object holding input parameters.


    Returns
    -------
    locations : pandas dataframe
        holding the well locations of unambiguous mutants

    """
    dims = experiment.pool_dims
    
    pool_reads = data[experiment.pools].astype(int)
        
    orientation = data["Orientation"]
    coords = data["Feature"]
    # get locations of unambiguous mutants in same format as predicted
    
    
    if "Barcode" in data.columns:
        outcols = ["Feature", "Orientation", "Barcode",
                  "Predicted_wells", "Possible_wells", "Ambiguity"]
        barcode = True
        barcodes = data["Barcode"]
    else:
        barcode = False
        outcols = ["Feature", "Orientation",
                  "Predicted_wells", "Possible_wells", "Ambiguity"]

    locations = pd.DataFrame(index=range(pool_reads.shape[0]), 
                                         columns=outcols)
    for i in range(pool_reads.shape[0]):
        line = pool_reads.iloc[i,:]
        nonzero_pools = []
        for j in dims:
            nonzero_pools.append(line[dims[j]][line[dims[j]]> 0].index.tolist())
        adds =  product(*nonzero_pools)

        wells = []
        for a in adds:
            wells.append("_".join(a))
        locations.loc[i,"Predicted_wells"] = ";".join(wells)
        locations.loc[i,"Possible_wells"] = ";".join(wells)
        locations.loc[i,"Orientation"] = orientation.values[i]
        locations.loc[i, "Feature"] =  coords.values[i]
        locations.loc[i, "Ambiguity"] =  "Unambiguous"
        if barcode == True:
            locations.loc[i, "Barcode"] =  barcodes.values[i]
    
    return locations


def coords2genes(experiment, locations):
    """
    
    Infer genes from genomic coordinates using input genbank file.

    Parameters
    ----------
    experiment :  LibraryExperiment
        Object holding input parameters.
    locations : pandas dataframe
        holding the well locations the tn mutants


    Returns
    -------
    out : pandas dataframe
        output table holding mutant information and well locations

    """
    data = experiment.count_mat
    # reference genbank file
    recs = [rec for rec in SeqIO.parse(experiment.gb_ref, "genbank")]
    # # - means complement strand -> reverse start end
    feats = [feat for feat in recs[0].features if feat.type == "CDS"]
    
    pools = experiment.pools
    
    outcols = ["Coordinate", "Gene", "Locus_tag", 
                                               "Predicted_wells","Gene_orientation",
                                               "Transposon_orientation",
                                               "Percent_from_start", "Gene_hit", 
                                               "Ambiguity", "Possible_wells", "Pool_counts"]
    out = pd.DataFrame(index=range(data.shape[0]), columns = outcols)
    
    barcode = False
    if "Barcode" in locations.columns:
        barcode = True
        out["Barcode"] = data["Barcode"].reset_index(drop=True)


    # store nonzero pool counts in final summary table
    pool_reads = data[pools].astype(float)
    for i in range(pool_reads.shape[0]):
        line = pool_reads.iloc[i,:]
        line = line[line > 0]
        pool_counts = [p+":" + str(i) for p,i in zip(line.index, line)]   
        pool_counts = ";".join(pool_counts      )    
        out.loc[i,"Pool_counts"] = pool_counts
        
    coords = data["Feature"]
    orientation = data["Orientation"]
    # populate with coords and orientation from pool presence table
    out["Coordinate"]  = coords.values
    out["Transposon_orientation"] = orientation.values
    # fill output dataframe, for now only with unambiguous mutants
    # this dataframe is equivalent to the progenitor_table of kosudoku, but ordered by mutants not wells
    for  i in range(locations.shape[0]):
        found = False
        c  = locations["Feature"].values[i]
        orientation_ = locations["Orientation"].values[i]
        # get idx of correct coordinate orientation pair
        if barcode == True:
            bar = locations["Barcode"].values[i]
            idx = np.where(np.logical_and(np.logical_and( out["Coordinate"] ==c , out["Transposon_orientation"] ==orientation_), out["Barcode"]==bar))[0]
        else:
            idx = np.where(np.logical_and( out["Coordinate"] ==c, out["Transposon_orientation"] ==orientation_))[0]
        c = int(c)
        for f in feats:
            start = int(f.location.start)
            end = int(f.location.end)
            if (start < c < end):
                found = True
            if found == True:
                qual = f.qualifiers
                if "gene" in qual:
                    out.loc[idx,"Gene"] = ",".join(qual["gene"] )
                else:
                    out.loc[idx,"Gene"] =  "Not_found"
    
                if "locus_tag" in qual:
                    out.loc[idx, "Locus_tag"] = ",".join(qual["locus_tag"])
                else:
                    out.loc[idx,"Locus_tag"] =  "Not_found"
    
                if f.strand == 1:
                    length = end-start
                    length_from_start = c-start
                    percent_from_start = length_from_start/length
                    out.loc[idx, "Percent_from_start"] = percent_from_start
                    out.loc[idx, "Gene_orientation"] = "+"
    
                if f.strand == -1:
                    length = end-start
                    length_from_start = end - c
                    percent_from_start = length_from_start/length
                    out.loc[idx, "Percent_from_start"] = percent_from_start
                    out.loc[idx, "Gene_orientation"] = "-"
    
                break
        out.loc[idx, "Predicted_wells"] = locations["Predicted_wells"].values[i]
        out.loc[idx, "Possible_wells"] = locations["Possible_wells"].values[i]
        out.loc[idx, "Ambiguity"] = locations["Ambiguity"].values[i]
        out.loc[idx, "Transposon_orientation"] = orientation_
        if found: 
           out.loc[idx, "Gene_hit"] = True
        else:
           out.loc[idx, "Gene_hit"] = False 

        
    
    return out

def transpose_location_summary(experiment, summary):
    """
    Transpose mutant summary  to well summary where each row corresponds to 
    a different well.

    Parameters
    ----------
    experiment : LibraryExperiment
        Object holding input parameters.
    summary : pandas dataframe
        output table holding mutant information and well locations

    Returns
    -------
    out : pandas dataframe
       well summary

    """
    def append2cell(cell, val):
        if cell == "nan":
            out = val
        else:
            out = ";".join([str(cell),str(val)])
        return out

    
    dims = experiment.pool_dims
    
    well_pools = list(product(*dims.values()))
    
    wells = np.array(["_".join(w) for w in well_pools])
    wells = wells.reshape(len(well_pools),1)
    well_pools = np.array(well_pools)
    
    incols = ["Well"] +list(dims.keys()) + ["Coordinate", "Gene", "Locus_tag", 
                                              "Gene_orientation", "Transposon_orientation",
                                               "Percent_from_start", "Gene_hit", 
                                               "Ambiguity"]
    outcols = ["Coordinate", "Gene", "Locus_tag", "Gene_orientation", 
               "Transposon_orientation","Percent_from_start", "Gene_hit", "Ambiguity"]
    barcode = False
    if "Barcode" in summary.columns:
        barcode = True
        dummy = np.empty((well_pools.shape[0],9))
        dummy[:] = np.nan
        incols = incols + ["Barcode"]
        outcols = outcols + ["Barcode"]
    else:
        dummy = np.empty((well_pools.shape[0],8))
        dummy[:] = np.nan
    grid = np.hstack([wells, well_pools, dummy])

    out = pd.DataFrame(grid, columns= incols)
    out.index = out["Well"]

    for row  in summary.itertuples(index =False):
        w = row[3]
        if type(w) == float and  math.isnan(w):
            continue
        else:
            w = w.split(";")
            c = row[0]
            g = row[1]
            l = row[2]
            gorient = row[4]
            torient = row[5]
            pfromstart = row[6]
            ghit = row[7]
            amb = row[8]
            if barcode==True:
                bar = row[11]
                outvals = [c,g,l,gorient,torient,pfromstart,ghit,amb,bar]
            else:
                outvals = [c,g,l,gorient,torient,pfromstart,ghit,amb]
            for i in w:
                for j,v in zip(outcols,outvals):
                    out.loc[i,j] = append2cell(out.loc[i,j], v)
    
    return out
    
    
    
